"""
Batch simulation runner
"""

from typing import Optional, Union

import chex
import jax

from .env import Sim, TSimParams


def batch_sim_runner(
    sim: Sim,
    n_samples: int,
    n_steps: int,
    rng: Union[chex.PRNGKey, int],
    show_progress: bool = True,
    params: Optional[TSimParams] = None,
    param_samples: Optional[TSimParams] = None,
) -> chex.ArrayTree:
    """
    Batch Monte-Carlo and parameter sweep simulation execution

    Run simulations across multiple random keys and optionally
    across a sample of simulation parameters.

    Parameters
    ----------
    sim: Sim
        Simulation functionality class.
    n_steps: int
        Number of simulation steps.
    n_samples: int
        Number of Monte-Carlo samples.
    rng: int or jax.random.PRNGKey
        Either a JAX random key or integer random seed.
    show_progress: bool, optional
        If ``True`` simulation progress bar will be displayed.
        Default ``True``.
    params: TSimParams, optional
        Optional simulation parameters. If not provided default
        simulation parameters will be used.
    param_samples: TSimParams, optional
        Optional simulation parameter samples to generate
        data across. Should have the same tree structure as individual
        simulation parameters.

    Returns
    -------
    Any
        Simulation records, with shape ``[n_samples, n_steps]`` if run
        with a single set of parameters, or shape ``[n_params, n_samples, n_steps]``
        if a set of parameters is provided.
    """
    key = rng if isinstance(rng, chex.PRNGKey) else jax.random.PRNGKey(rng)

    assert (
        params is None or param_samples is None
    ), "Either params or param_samples should be provided, not both"

    if params is None and param_samples is None:
        params = sim.default_params()

    def inner(k, _params):
        k1, k2 = jax.random.split(k)
        _initial_state = sim.initial_state(k1)
        _, records, _ = sim.run(
            n_steps, k2, _params, _initial_state, show_progress=show_progress
        )
        return records

    def sample_params(k, _params):
        keys = jax.random.split(k, n_samples)
        return jax.vmap(inner, in_axes=(0, None))(keys, _params)

    if params is not None:
        batch_records = sample_params(key, params)
    else:
        batch_records = jax.vmap(sample_params, in_axes=(None, 0))(key, param_samples)

    return batch_records



"""
Evosax simulation update utilities
"""
from functools import partial
from typing import Callable

import chex
import jax


@partial(jax.jit, static_argnames=("f",))
def broadcast_params(
    f: Callable, params: chex.ArrayTree, obs: chex.ArrayTree
) -> chex.ArrayTree:
    """
    Broadcast evosax parameters over observations

    Applies parameters where agents all share
    the same parameters, across their individual
    observations, returning an output for
    each agent.

    Parameters
    ----------
    f: Callable
        Function parameterised by ``params`` (e.g.
        a neural-network apply function).
    params: chex.ArrayTree
        Evosax function parameters. Since shared
        by agents these should be a single sample
        of parameters from a larger population.
    obs: chex.ArrayTree
        Array/tree of individual observation for agents
        to be mapped over.

    Returns
    -------
    chex.ArrayTree
        Output of ``f`` for each agent
    """
    return jax.vmap(f, in_axes=(None, 0))(params, obs)


@partial(jax.jit, static_argnames=("f",))
def map_params(
    f: Callable, params: chex.ArrayTree, obs: chex.ArrayTree
) -> chex.ArrayTree:
    """
    Map an evosax population of parameters over observations

    Applies a population of parameters across agents
    individual observations, returning an output for
    each agent.

    Parameters
    ----------
    f: Callable
        Function parameterised by ``params`` (e.g.
        a neural-network apply function).
    params: chex.ArrayTree
        Population of Evosax function parameters.
    obs: chex.ArrayTree
        Array/tree of individual observation for agents
        to be mapped over.

    Returns
    -------
    chex.ArrayTree
        Output of ``f`` for each agent
    """
    return jax.vmap(f, in_axes=(0, 0))(params, obs)



from functools import partial
from typing import Any, Callable, Optional, Tuple

import chex
import jax
import jax.numpy as jnp

from esquilax import utils


def spatial(
    n_bins: int,
    reduction: chex.ArrayTree,
    default: chex.ArrayTree,
    include_self: bool = False,
    topology: str = "moore",
    i_range: Optional[float] = None,
) -> Callable:
    """
    Apply a function between agents based on spatial proximity

    This transformation efficiently checks
    for spatial proximity between agents before applying
    the interaction function to pairs of agents
    below a threshold range of each other.

    .. warning::

       This implementation currently assumes a 2-dimensional
       space with continues boundary conditions (i.e. wrapped
       on a torus).

    .. note::

       Performance is dependent on the proximity threshold
       and density of agents. Worse case it will perform
       :math:`N^2` checks if all agents are in the same
       neighbourhood.

    Examples
    --------

    This example subdivides the space into 2 cells along
    each dimension (i.e. there are 4 cells total),
    then sums contributions from neighbours, excluding the
    active agent

    .. testsetup:: spatial

       import esquilax
       import jax
       import jax.numpy as jnp

    .. testcode:: spatial

       @esquilax.transforms.spatial(
           2, jnp.add, 0, include_self=False, topology="same-cell"
       )
       def foo(_k, p, a, b):
           return p + a + b

       x = jnp.array([[0.1, 0.1], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(3)
       k = jax.random.PRNGKey(101)

       foo(k, 2, a, a, pos=x)
       # [0, 5, 5]

    .. doctest:: spatial
       :hide:

       >>> foo(k, 2, a, a, pos=x).tolist()
       [0, 5, 5]

    so in this case agent ``0`` does not have any neighbours in
    its cell, but agents ``1`` and ``2`` observe each other.

    Arguments and return values can be PyTrees or multidimensional
    arrays. Arguments can also be ``None`` if not used

    .. testcode:: spatial

       @esquilax.transforms.spatial(
           2,
           (jnp.add, jnp.add),
           (0, 0),
           include_self=False,
           topology="same-cell",
       )
       def foo(_k, p, _, b):
           return p + b[0],  p + b[1]

       x = jnp.array([[0.1, 0.1], [0.6, 0.6], [0.7, 0.7]])
       a = jnp.arange(6).reshape(3, 2)
       k = jax.random.PRNGKey(101)

       foo(k, 2, None, a, pos=x)
       # ([0, 6, 4], [0, 7, 5])

    .. doctest:: spatial
       :hide:

       >>> foo(k, 2, None, a, pos=x)
       (Array([0, 6, 4], dtype=int32), Array([0, 7, 5], dtype=int32))

    Parameters
    ----------
    n_bins: int
        Number of bins each dimension is subdivided
        into. Assumes that each dimension contains the
        same number of cells. Each cell can only interact
        with adjacent cells, so this value also consequently
        also controls the number of interactions.
    reduction: Callable
        Binary monoidal reduction function, eg ``jax.numpy.add``.
    default: Any
        Default/identity reduction value
    include_self: bool, optional
        if ``True`` each agent will include itself in the
        gathered values.
    topology: str, optional
        Topology of cells, default ``"moore"``. Since cells
        interact with their neighbours, topologies with
        fewer neighbours can increase performance at the
        cost of fidelity. Should be one of ``"same-cell"``,
        ``"von-neumann"`` or ``"moore"``.
    i_range: float, optional
        Optional interaction range. By default, the width
        of a cell is used as the interaction range, but this
        can be increased/decreased using ``i_range`` dependent
        on the use-case.
    f: Callable
        Interaction to apply to in-proximity pairs, should
        have the signature

        .. code-block:: python

           def f(
               k: chex.PRNGKey,
               params: Any,
               a: Any,
               b: Any,
               **static_kwargs,
           ):
               ...
               return x

        where

        - ``k``: Is a JAX random key
        - ``params``: Parameters broadcast over all interactions
        - ``a``: Start agent in the interaction
        - ``b``: End agent in the interaction
        - ``**static_kwargs``: Any arguments required at compile
          time by JAX can be passed as keyword arguments.
    """
    w = 1.0 / n_bins
    i_range = w if i_range is None else i_range

    chex.assert_trees_all_equal_structs(
        reduction, default
    ), "Reduction and default PyTrees should have the same structure"

    def spatial_decorator(f: Callable) -> Callable:
        t = utils.space.get_cell_neighbours(n_bins, topology)
        keyword_args = utils.functions.get_keyword_args(f)

        @partial(jax.jit, static_argnames=keyword_args)
        def _spatial(
            key: chex.PRNGKey,
            params: Any,
            agents_a: Any,
            agents_b: Any,
            *,
            pos: chex.Array,
            **static_kwargs,
        ) -> Any:
            idxs = utils.space.get_bins(pos, n_bins, w)
            sort_idxs = jnp.argsort(idxs)

            _, bins = utils.graph.index_bins(idxs, n_bins**2)

            sorted_idxs = idxs[sort_idxs]
            sorted_x = pos[sort_idxs]
            sorted_agents_a = jax.tree_util.tree_map(lambda y: y[sort_idxs], agents_a)
            sorted_agents_b = jax.tree_util.tree_map(lambda y: y[sort_idxs], agents_b)

            def cell(k: chex.PRNGKey, i: int, bin_range: chex.Array) -> Any:
                agent_a = jax.tree_util.tree_map(lambda y: y[i], sorted_agents_a)
                pos_a = sorted_x[i]

                def interact(
                    j: int, _k: chex.PRNGKey, _r: Any
                ) -> Tuple[chex.PRNGKey, Any]:
                    _k, fk = jax.random.split(_k, 2)
                    agent_b = jax.tree_util.tree_map(lambda z: z[j], sorted_agents_b)
                    r = partial(f, **static_kwargs)(fk, params, agent_a, agent_b)
                    r = jax.tree_util.tree_map(
                        lambda red, a, b: red(a, b), reduction, _r, r
                    )
                    return _k, r

                def inner(
                    j: int, carry: Tuple[chex.PRNGKey, Any]
                ) -> Tuple[chex.PRNGKey, Any]:
                    _k, _r = carry
                    pos_b = sorted_x[j]
                    d = utils.space.shortest_distance(pos_a, pos_b, 1.0, norm=False)
                    return jax.lax.cond(
                        d < i_range, interact, lambda _, _x, _z: (_x, _z), j, _k, _r
                    )

                if include_self:
                    _, _results = jax.lax.fori_loop(
                        bin_range[0], bin_range[1], inner, (k, default)
                    )
                else:
                    k, _results = jax.lax.fori_loop(
                        bin_range[0],
                        jnp.minimum(i, bin_range[1]),
                        inner,
                        (k, default),
                    )
                    _, _results = jax.lax.fori_loop(
                        jnp.maximum(i + 1, bin_range[0]),
                        bin_range[1],
                        inner,
                        (k, _results),
                    )

                return _results

            def agent_reduce(k: chex.PRNGKey, i: int, bin_idx: int):
                nbs = t[bin_idx]
                nb_bins = bins[nbs]
                _keys = jax.random.split(k, nbs.shape[0])
                _results = jax.vmap(cell, in_axes=(0, None, 0))(_keys, i, nb_bins)

                def red(a, _, c):
                    return jnp.frompyfunc(c, 2, 1).reduce(a)

                return jax.tree_util.tree_map(red, _results, default, reduction)

            n_agents = pos.shape[0]
            keys = jax.random.split(key, n_agents)
            results = jax.vmap(agent_reduce, in_axes=(0, 0, 0))(
                keys, jnp.arange(n_agents), sorted_idxs
            )

            inv_sort = jnp.argsort(sort_idxs)
            results = jax.tree_util.tree_map(lambda y: y[inv_sort], results)

            return results

        return _spatial

    return spatial_decorator



"""
Wrappers for `Evosax`_ strategy functionality

.. _Evosax: https://github.com/RobertTLange/evosax?tab=readme-ov-file
"""
from functools import partial
from typing import Any, Dict, Tuple, Union

import chex
import evosax
import jax
from flax.typing import FrozenVariableDict


class Strategy:
    """
    Abstract class wrapping Evosax strategy functionality

    This class can be used to encapsulate multiple populations,
    for example to have multiple agent types parameterised
    by independent populations.

    .. note::

       In the case of multiple policies, the tree structure
       should be consistent across population states,
       population samples etc.
    """

    def initialize(
        self, k: chex.PRNGKey, evo_params: evosax.EvoParams
    ) -> evosax.EvoState:
        """
        Initialise strategy state/states

        Parameters
        ----------
        k: jax.random.PRNGKey
            JAX random key
        evo_params: evosax.EvoParams
            Strategy parameters

        Returns
        -------
        evosax.EvoState
            Strategy state or collection of states
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def default_params(self) -> evosax.EvoParams:
        """
        Get default strategy parameters

        Returns
        -------
        evosax.EvoParams
            Strategy parameters
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def reshape_params(self, population) -> chex.ArrayTree:
        """
        Reshape parameters

        Parameters
        ----------
        population
            Array (or collection of arrays) of parameters
            sampled from the strategy/strategies.

        Returns
        -------
        Reshaped parameters or collection of parameters
        """
        return population

    @partial(jax.jit, static_argnums=(0,))
    def shape_rewards(self, population, fitness: chex.ArrayTree) -> chex.ArrayTree:
        """
        Reshape environment rewards

        Rescale/reshape rewards generated during a simulation.

        Parameters
        ----------
        population
            Strategy population or collection of populations
        fitness: jax.numpy.array
            Fitness/rewards (or collection of)

        Returns
        -------
        chex.ArrayTree
            Rescaled fitness
        """
        return fitness

    @partial(jax.jit, static_argnums=(0,))
    def ask(
        self, k: chex.PRNGKey, evo_state: evosax.EvoState, evo_params: evosax.EvoParams
    ) -> Tuple[chex.ArrayTree, evosax.EvoState]:
        """
        Sample parameters from the current strategy state

        Parameters
        ----------
        k: jax.random.PRNGKey
            JAX random key
        evo_state: evosax.EvoState
            Strategy state (or collection of states)
        evo_params: evosax.EvoParams
            Strategy parameters (or collection of parameters)

        Returns
        -------
        tuple
            Population array (or collection of arrays) and
            updated state of the strategy/strategies
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        population: chex.ArrayTree,
        fitness: chex.ArrayTree,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> evosax.EvoState:
        """
        Update strategy state/states

        Parameters
        ----------
        population: chex.ArrayTree
            Population array/arrays
        fitness: chex.ArrayTree
            Fitness array/arrays
        evo_state: evosax.EvoState
            Strategy state or collection of states
        evo_params: evosax.EvoParams
            Strategy params or collection of params

        Returns
        -------
        evosax.EvoState
            Update state/states
        """
        raise NotImplementedError


class BasicStrategy(Strategy):
    """
    Basic strategy representing a single policy

    Wrapper around a single strategy, with parameters
    initialised from a Flax neural-network.
    """

    def __init__(
        self,
        network_params: Union[FrozenVariableDict, Dict[str, Any]],
        strategy,
        pop_size: int,
        centered_rank_fitness: bool = True,
        z_score_fitness: bool = False,
        maximize_fitness: bool = True,
        **strategy_kwargs,
    ):
        """
        Parameters
        ----------
        network_params: Union[FrozenVariableDict, Dict[str, Any]]
            Flax network parameters.
        strategy
            Evosax strategy class.
        pop_size: int
            Strategy population size.
        centered_rank_fitness: bool, optional
            Use ``centered_rank_fitness`` for fitness-shaping, default ``True``.
        z_score_fitness: bool, optional
            Use ``z_score_fitness`` for fitness-shaping, default ``False``.
        maximize_fitness: bool, optional
            Use ``maximize_fitness`` for fitness-shaping, default ``True``.
        **strategy_kwargs
            Keyword arguments to pass to the strategy constructor.

        Attributes
        ----------
        param_reshaper: evosax.ParameterReshaper
            Parameter reshaper function
        strategy
            Initialised strategy
        fitness_shaper: evosax.FitnessShaper
            Fitness rescaler
        """
        self.param_reshaper = evosax.ParameterReshaper(network_params)
        self.strategy = strategy(
            popsize=pop_size,
            num_dims=self.param_reshaper.total_params,
            **strategy_kwargs,
        )
        self.fitness_shaper = evosax.FitnessShaper(
            centered_rank=centered_rank_fitness,
            z_score=z_score_fitness,
            maximize=maximize_fitness,
        )

    def initialize(
        self, k: chex.PRNGKey, evo_params: evosax.EvoParams
    ) -> evosax.EvoState:
        """
        Initialise strategy state

        Parameters
        ----------
        k: jax.random.PRNGKey
            JAX random key
        evo_params: evosax.EvoParams
            Strategy parameters

        Returns
        -------
        evosax.EvoState
            Strategy state
        """
        return self.strategy.initialize(k, evo_params)

    @partial(jax.jit, static_argnums=(0,))
    def default_params(self) -> evosax.EvoParams:
        """
        Get default strategy parameters

        Returns
        -------
        evosax.EvoParams
            Strategy parameters
        """
        return self.strategy.default_params

    def reshape_params(self, population) -> chex.ArrayTree:
        """
        Reshape parameters

        Parameters
        ----------
        population: jax.numpy.array
            Array of parameters sampled from the strategy.

        Returns
        -------
        jax.numpy.array
            Rescaled parameters
        """
        return self.param_reshaper.reshape(population)

    def ask(
        self, k: chex.PRNGKey, evo_state: evosax.EvoState, evo_params: evosax.EvoParams
    ) -> Tuple[chex.ArrayTree, evosax.EvoState]:
        """
        Sample parameters from the current strategy state

        Parameters
        ----------
        k: jax.random.PRNGKey
            JAX random key
        evo_state: evosax.EvoState
            Strategy state
        evo_params: evosax.EvoParams
            Strategy parameters

        Returns
        -------
        tuple[jax.numpy.array, evosax.EvoState]
            Population array and updated state of the strategy
        """
        return self.strategy.ask_strategy(k, evo_state, evo_params)

    def shape_rewards(self, population, fitness: chex.ArrayTree) -> chex.ArrayTree:
        """
        Reshape rewards

        Rescale/reshape rewards generated during a simulation.

        Parameters
        ----------
        population
            Strategy population or collection of populations
        fitness: jax.numpy.array
            Fitness/rewards (or collection of)

        Returns
        -------
        chex.ArrayTree
            Rescaled fitness
        """
        return self.fitness_shaper.apply(population, fitness)

    def tell(
        self,
        population: chex.ArrayTree,
        fitness: chex.ArrayTree,
        evo_state: evosax.EvoState,
        evo_params: evosax.EvoParams,
    ) -> evosax.EvoState:
        """
        Update strategy state

        Parameters
        ----------
        population: chex.ArrayTree
            Population array
        fitness: chex.ArrayTree
            Fitness array
        evo_state: evosax.EvoState
            Strategy state
        evo_params: evosax.EvoParams
            Strategy params

        Returns
        -------
        evosax.EvoState
            Update state
        """
        return self.strategy.tell(population, fitness, evo_state, evo_params)



"""
`Evosax`_ policy training and testing functionality

.. _Evosax: https://github.com/RobertTLange/evosax?tab=readme-ov-file
"""
from functools import partial
from typing import Optional, Tuple

import chex
import evosax
import jax
import jax.numpy as jnp
import jax_tqdm

from esquilax.batch_runner import batch_sim_runner
from esquilax.env import Sim, TSimParams

from .strategy import Strategy


@chex.dataclass
class TrainingData:
    """
    Dataclass tracking training rewards and recorded data

    Simulations used for training should return
    a ``TrainingData`` class as its recorded data.

    Examples
    --------

    .. code-block:: python

       def step(self, i, k, params, state):
           ...
           # Rewards are returned as part of the TrainingData
           return (
               state,
               TrainingData(rewards=rewards, records=records)
           )

    Parameters
    ----------
    rewards: chex.ArrayTree
        PyTree of rewards generated by simulation agents.
    records: chex.ArrayTree
        Any additional data recorded over the course
        of the simulation.
    """

    rewards: chex.Array
    records: chex.ArrayTree


@partial(
    jax.jit,
    static_argnames=(
        "strategy",
        "env",
        "n_generations",
        "n_samples",
        "n_steps",
        "map_population",
        "show_progress",
    ),
)
def train(
    strategy: Strategy,
    env: Sim,
    n_generations: int,
    n_samples: int,
    n_steps: int,
    map_population: bool,
    k: chex.PRNGKey,
    evo_params: evosax.EvoParams,
    evo_state: evosax.EvoState,
    show_progress: bool = True,
    env_params: Optional[TSimParams] = None,
) -> Tuple[evosax.EvoState, chex.ArrayTree]:
    """
    Train a policy/policies using neuro-evolution

    Training loop that optimises a policy evaluated
    against an Esquilax simulation training
    environment.

    Agent policies/behaviours are sampled and updated
    using Evosax, and can either be tested as
    a shard policy (i.e. all agents share a policy
    from the population), or each agent is controlled
    by an individual member of the population.

    The training loop broadly follows these steps:

    - Sample a new population from the strategy
    - Initialise a simulation(s) and run with
      the population/parameters
    - Collect rewards over the simulation
    - Aggregate and rescale rewards
    - Update the strategy state from the
      rewards/fitness

    .. note::

       Rewards are summed over the simulation steps, i.e.
       the total rewards for each agent (or policy)
       are used to measure fitness during training.

    Parameters
    ----------
    strategy: esquilax.evo.Strategy
        Evosax evolutionary strategy
    env: esquilax.env.SimEnv
        Esquilax simulation training environment

        .. warning::

           The step function of the environment should
           include the population/parameters as
           a member of the parameters argument, and
           return a `TrainingData` class, for example:

           .. code-block:: python

              def step(self, i, k, params, state):
                  # params is a tuple containing the population
                  # as the first member
                  population, sim_params = params
                  ...
                  # Rewards are returned as part of the TrainingData
                  return (
                      state,
                      TrainingData(rewards=rewards, records=records)
                  )

           The current population and rewards will then be
           passed and handled automatically by the training
           loop.
    n_generations: int
        Number of training generations
    n_samples: int
        Number of Monte-Carlo samples to test each population
        or parameter samples across.
    n_steps: int
        Number of steps to run each simulation
    map_population: bool
        If ``True`` each member of the population is
        evaluated in a separate environment (i.e. the
        individual parameters are shared by agents). If
        ``False`` the whole population is passed
        to the environment as part of the simulation
        state.
    k: jax.random.PRNGKey
        JAX random key
    evo_params: evosax.EvoParams
        Evosax strategy parameters
    evo_state: evosax.EvoState
        Evosax strategy state
    show_progress: bool, optional
        If ``True`` a progress bar will be displayed
        showing training progress. Default value is ``True``.
    env_params: TSimParams, optional
        Optional environment parameters, if not provided the
        default environment parameters will be used.

    Returns
    -------
    tuple[evosax.EvoState, chex.ArrayTree]
        Trained strategy state and record of rewards
        gathered over the course of training.
    """

    env_params = env.default_params() if env_params is None else env_params

    def generation(carry, _):
        _k, _evo_state = carry

        _k, k1, k2 = jax.random.split(_k, 3)

        _population, _evo_state = strategy.ask(k1, _evo_state, evo_params)
        _population_shaped = strategy.reshape_params(_population)

        def inner(_pop) -> TrainingData:
            _params = (_pop, env_params)
            return batch_sim_runner(
                env, n_samples, n_steps, k2, show_progress=False, params=_params
            )

        if map_population:
            _training_data = jax.vmap(inner)(_population_shaped)
            _total_rewards = jnp.sum(_training_data.rewards, axis=(0, 2, 3))
        else:
            _training_data = inner(_population_shaped)
            _total_rewards = jnp.sum(_training_data.rewards, axis=(0, 1))

        _fitness = strategy.shape_rewards(_population, _total_rewards)
        _evo_state = strategy.tell(_population, _fitness, _evo_state, evo_params)

        return (_k, _evo_state), _total_rewards

    if show_progress:
        generation = jax_tqdm.scan_tqdm(n_generations, desc="Generation")(generation)

    (k, evo_state), rewards = jax.lax.scan(
        generation,
        (k, evo_state),
        jnp.arange(n_generations),
    )

    return evo_state, rewards


@partial(
    jax.jit,
    static_argnames=(
        "env",
        "n_steps",
        "map_population",
        "show_progress",
    ),
)
def test(
    population_shaped: chex.ArrayTree,
    env: Sim,
    n_steps: int,
    map_population: bool,
    k: chex.PRNGKey,
    show_progress: bool = True,
    env_params: Optional[TSimParams] = None,
) -> tuple[chex.ArrayTree, chex.ArrayTree]:
    """
    Test population performance and gather telemetry

    Test a population against a simulation environment
    gathering rewards and recorded simulation state
    over the course of the simulation

    Parameters
    ----------
    population_shaped: chex.ArrayTree
        Reshaped population parameters for use
        by simulation agents.
    env: esquilax.env.SimEnv
        Esquilax simulation training environment
    n_steps: int
        Number of simulation steps.
    map_population: bool
        If ``True`` each member of the population is
        evaluated in a separate environment (i.e. the
        individual parameters are shared by agents). If
        ``False`` the whole population is passed
        to the environment as part of the simulation
        state.
    k: jax.random.PRNGKey
        JAX random key
    show_progress: bool, optional
        If ``True`` a progress bar will be displayed
        showing simulation progress. Default value is ``True``.
    env_params: TSimParams, optional
        Optional environment parameters, if not provided the
        default environment parameters will be used.

    Returns
    -------
    tuple[chex.ArrayTree, chex.ArrayTree]
        Tuple containing records of rewards and recorded
        state for each step of the simulation.
    """
    k1, k2 = jax.random.split(k)

    env_params = env.default_params() if env_params is None else env_params

    def inner(_pop):
        _initial_state = env.initial_state(k1)
        _params = (_pop, env_params)
        _, _testing_data, _ = env.run(
            n_steps, k2, _params, _initial_state, show_progress=show_progress
        )
        return _testing_data

    if map_population:
        records = jax.vmap(inner)(population_shaped)
    else:
        records = inner(population_shaped)

    return records



"""
Policy training patterns for use with `Evosax`_ neuro-evolution package

Base-classes and functionality for training of evolutionary
policies (via `Evosax`_) using esquilax as a multi-agent
(and potentially multi-policy) training environment.

.. _Evosax: https://github.com/RobertTLange/evosax?tab=readme-ov-file
"""
from .strategy import BasicStrategy, Strategy
from .training import TrainingData, test, train
from .updates import broadcast_params, map_params



"""
Functions used in spatial transforms
"""
from itertools import product
from typing import Union

import chex
import jax
import jax.numpy as jnp


def get_bins(x: chex.Array, n: int, w: float) -> chex.Array:
    """
    Assign co-ordinates to a grid-cell

    .. warning::

       This implementation currently assumes that the width
       of cells and number of cells is the same in both
       dimensions.

    Parameters
    ----------
    x: jax.ndarray
        2d array of co-ordinates, in shape ``[n, 2]``.
    n: int
        Number of cells along dimensions.
    w: float
        Width of cells.

    Returns
    -------
    jax.ndarray
        Array of cell indices for each position
    """
    y = jnp.floor_divide(x, w).astype(jnp.int32)
    i = y[:, 0] * n + y[:, 1]
    return i


def get_cell_neighbours(n_bins: int, topology: str) -> chex.Array:
    """
    Get indices of adjacent cells for each cell on a grid

    Returns array where each row represents a cell
    and indices of cells adjacent to it given
    a desired topology.

    .. warning::

       This implementation currently assumes that the
       space is wrapped at the edges (i.e. on a torus).

    Parameters
    ----------
    n_bins: int
        Number of bins
    topology: str
        Topology of the neighbouring cells, one of:

        - ``same-cell``: Only consider cells in isolation
        - ``von-neumann``: Use a `Von-Neumann neighbourhood \
        <https://en.wikipedia.org/wiki/Von_Neumann_neighborhood>`_
        - ``moore``: Use the `Moore neighbourhood \
        <https://en.wikipedia.org/wiki/Moore_neighborhood>`_

    Returns
    -------
    jax.ndarray
        2d array where each row contains a cells index
        and indices of its neighbours.
    """
    n = n_bins**2

    if topology == "same-cell":
        offsets = jnp.array([[0, 0]])
    elif topology == "von-neumann":
        offsets = jnp.array([[0, 0], [0, 1], [1, 0], [0, -1], [-1, 0]])
    elif topology == "moore":
        offsets = list(product([-1, 0, 1], [-1, 0, 1]))
        offsets = jnp.array(offsets)
    else:
        raise ValueError(
            "Topology should be one of 'same-cell', 'von-neumann' or 'moore'"
        )

    grid = jnp.arange(n).reshape(n_bins, n_bins)
    id_x, id_y = jnp.meshgrid(jnp.arange(n_bins), jnp.arange(n_bins))

    def slice(off):
        off_x, off_y = id_x + off[1], id_y + off[0]
        off_x, off_y = off_x % n_bins, off_y % n_bins
        x = grid[off_y, off_x].reshape(-1)
        return x

    idxs = jax.vmap(slice, out_axes=1)(offsets)

    return idxs


def shortest_vector(a: chex.Array, b: chex.Array, length: float = 1.0) -> chex.Array:
    """
    Get the shortest vector between points on a torus

    Parameters
    ----------
    a: jax.ndarray
        Co-ordinate vector.
    b: jax.ndarray
        Co-ordinate vector.
    length: float, optional
        Length of the space. Default ``1.0``.

    Returns
    -------
    jax.ndarray
        Array representing the shortest vector between the points.
    """
    x = b - a
    x_ = jnp.sign(x) * (jnp.abs(x) - length)
    return jnp.where(jnp.abs(x) < jnp.abs(x_), x, x_)


def shortest_distance(
    a: Union[float, chex.Array],
    b: Union[float, chex.Array],
    length: float = 1.0,
    norm: bool = True,
) -> Union[float, chex.Array]:
    """
    Get the shortest distance between points on a torus

    Parameters
    ----------
    a: jax.ndarray
        Co-ordinate vector
    b: jax.ndarray
        Co-ordinate vector
    length: float, optional
        Length of the space. Default is ``1.0``.
    norm: bool, optional
        If ``True`` the distance will be normalised
        otherwise the square of the distance will be
        returned. Default ``True``.

    Returns
    -------
    float
        Distance between the points
    """
    x = jnp.abs(a - b)
    d = jnp.where(x > 0.5 * length, length - x, x)
    d = jnp.sum(jnp.square(d), axis=-1)
    if norm:
        d = jnp.sqrt(d)
    return d



"""
Function transformations representing agent interactions and updates

The transformation implement map-reduce patterns that represent
agents/entities observing other agents/entities, and functions
that update the state of agents/entities.
"""
from ._graph import edge_map, graph_reduce, random_neighbour
from ._map import amap
from ._space import spatial



"""
Graph transformations

Transformations representing interactions between
nodes (and edges) on a graph.
"""
from functools import partial
from typing import Any, Callable

import chex
import jax
import jax.numpy as jnp

from esquilax import utils


def edge_map(f: Callable) -> Callable:
    """
    Map a function over graph edges and related nodes

    Maps a function over a set of edges (with data) and
    data corresponding to the edge start and end nodes,

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: edge_map

       import esquilax
       import jax
       import jax.numpy as jnp

    .. testcode:: edge_map

       @esquilax.transforms.edge_map
       def f(k, params, start, end, edge):
           return params + start + end + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       starts = jnp.array([0, 1, 2])
       ends = jnp.array([0, 1, 2])

       # Call transform with edge indexes
       f(k, 2, starts, ends, edges, edge_idxs=edge_idxs)
       # [3, 5, 5]

    .. testcode:: edge_map
       :hide:

       z = f(k, 2, starts, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [3, 5, 5]

    Arguments can also be PyTrees or ``None`` if unused

    .. testcode:: edge_map

       @esquilax.transforms.edge_map
       def f(_k, _params, _start, end, edge):
           return end[0] + end[1] + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       # [2, 5, 2]

    .. testcode:: edge_map
       :hide:

       z = f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [2, 5, 2]

    Parameters
    ----------
    f: Callable
        Function with the signature

        .. code-block:: python

           def f(k, params, start, end, edge, **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``k``: JAX PRNGKey
        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.
    """
    keyword_args = utils.functions.get_keyword_args(f)

    @partial(jax.jit, static_argnames=keyword_args)
    def _edge_map(
        k: chex.PRNGKey,
        params: Any,
        starts: Any,
        ends: Any,
        edges: Any,
        *,
        edge_idxs: chex.Array,
        **static_kwargs,
    ) -> Any:
        n = edge_idxs.shape[1]
        keys = jax.random.split(k, n)
        starts = jax.tree_util.tree_map(
            lambda x: x.at[edge_idxs[0]].get(),
            starts,
        )
        ends = jax.tree_util.tree_map(
            lambda x: x.at[edge_idxs[1]].get(),
            ends,
        )

        return jax.vmap(partial(f, **static_kwargs), in_axes=(0, None, 0, 0, 0))(
            keys, params, starts, ends, edges
        )

    return _edge_map


def graph_reduce(
    reduction: chex.ArrayTree, default: chex.ArrayTree, n: int = -1
) -> Callable:
    """
    Map function over graph edges and reduce results to nodes

    Maps the update function over the graph edges and corresponding
    node data. Then aggregates the results back to the start nodes
    of the edge-set. Edges are treated as directional, with the
    first node observing the state of the second.

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: graph_reduce

       import esquilax
       import jax
       import jax.numpy as jnp

    .. testcode:: graph_reduce

       @esquilax.transforms.graph_reduce(jnp.add, 0)
       def f(k, params, start, end, edge):
           return params + start + end + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       starts = jnp.array([0, 1, 2])
       ends = jnp.array([0, 1, 2])

       # Call transform with edge indexes
       f(k, 2, starts, ends, edges, edge_idxs=edge_idxs)
       # [8, 5, 0]

    .. testcode:: graph_reduce
       :hide:

       z =  f(k, 2, starts, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [8, 5, 0]

    Arguments can also be PyTrees, or ``None`` if unused

    .. testcode:: graph_reduce

       @esquilax.transforms.graph_reduce(jnp.add, 0, n=3)
       def f(k, _params, _start, end, edge):
           return end[0] + end[1] + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       # [7, 2, 0]

    .. testcode:: graph_reduce
       :hide:

       z = f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [7, 2, 0]

    Parameters
    ----------
    reduction: Callable
        Binary monoidal reduction function
    default: Any
        Default/identity value result value
    n: int, optional
        Number of nodes, should be provided if start-node data is ``None``
    f: Callable
        Function with the signature

        .. code-block:: python

           def f(k, params, start, end, edge,  **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``k``: JAX PRNGKey
        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.
    """

    def _graph_reduce_decorator(f: Callable) -> Callable:
        chex.assert_trees_all_equal_structs(
            reduction, default
        ), "Reduction and default PyTrees should have the same structure"

        _edge_map = edge_map(f)
        keyword_args = utils.functions.get_keyword_args(f)

        @partial(jax.jit, static_argnames=keyword_args)
        def _graph_reduce(
            k: chex.PRNGKey,
            params: Any,
            starts: Any,
            ends: Any,
            edges: Any,
            *,
            edge_idxs: chex.Array,
            **kwargs,
        ) -> Any:
            n_results = utils.functions.get_size(starts) if n < 0 else n

            edge_results = _edge_map(
                k, params, starts, ends, edges, **kwargs, edge_idxs=edge_idxs
            )
            bin_counts, bins = utils.graph.index_bins(edge_idxs[0], n_results)
            bins = bins.reshape(-1)

            def reduce(r, x, d):
                x = jnp.frompyfunc(r, 2, 1).reduceat(x, bins, axis=0)[::2]
                x = jnp.where(bin_counts > 0, x, d)
                return x

            results = jax.tree_util.tree_map(reduce, reduction, edge_results, default)

            return results

        return _graph_reduce

    return _graph_reduce_decorator


def random_neighbour(default: Any, n: int = -1) -> Callable:
    """
    Apply function to random selected graph neighbours

    For each start node select a random neighbour on
    the graph (i.e. a random edge starting at that node)
    and apply the update function. Returns a
    default value if the node has no outgoing edges.

    .. warning::

       Edge indices and any associated data should be
       sorted using :py:meth:`esquilax.utils.sort_edges`

    Examples
    --------

    .. testsetup:: random_neighbour

       import esquilax
       import jax
       import jax.numpy as jnp

    .. testcode:: random_neighbour

       @esquilax.transforms.random_neighbour(0, n=3)
       def f(k, _params, _start, end, edge):
           return end[0] + end[1] + edge

       k = jax.random.PRNGKey(101)
       edge_idxs = jnp.array([[0, 0, 1], [1, 2, 0]])
       edges = jnp.array([0, 1, 2])
       ends = (jnp.array([0, 1, 2]), jnp.array([0, 1, 2]))

       # Call transform with edge indexes
       f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       # [5, 2, 0]

    .. testcode:: random_neighbour
       :hide:

       z = f(k, None, None, ends, edges, edge_idxs=edge_idxs)
       assert z.tolist() == [5, 2, 0]

    Parameters
    ----------
    default: Any
        Value returned when a node has no neighbours
    n: int, optional
        Number of start nodes, should be provided if
        start-node data is None
    f: Callable
        Function with the signature

        .. code-block:: python

           def f(k, params, start, end, edge, **static_kwargs):
               ...
               return x

        where the arguments are:

        - ``k``: JAX PRNGKey
        - ``params``: Parameters (shared over the map)
        - ``start``: Start node data
        - ``end``: End node data
        - ``edge``: Edge data
        - ``**static_kwargs``: Any values required at compile-time
          by JAX can be passed as keyword arguments.
    """

    def random_neighbour_decorator(f: Callable) -> Callable:
        keyword_args = utils.functions.get_keyword_args(f)

        @partial(jax.jit, static_argnames=keyword_args)
        def _random_neighbour(
            k: chex.PRNGKey,
            params: Any,
            starts: Any,
            ends: Any,
            edges: Any,
            *,
            edge_idxs: chex.Array,
            **static_kwargs,
        ):
            n_results = utils.functions.get_size(starts) if n < 0 else n
            bin_counts, bins = utils.graph.index_bins(edge_idxs[0], n_results)
            keys = jax.random.split(k, n_results)

            def sample(_k, i, a, b):
                k1, k2 = jax.random.split(_k)
                j = jax.random.randint(k1, (), a, b)
                end_idx = edge_idxs[1][j]
                start = jax.tree_util.tree_map(lambda x: x[i], starts)
                end = jax.tree_util.tree_map(lambda x: x[end_idx], ends)
                edge = jax.tree_util.tree_map(lambda x: x[j], edges)
                return partial(f, **static_kwargs)(k2, params, start, end, edge)

            def select(_k, i, count: int, bin: chex.Array) -> Any:
                return jax.lax.cond(
                    count > 0, sample, lambda *_: default, _k, i, bin[0], bin[1]
                )

            return jax.vmap(select, in_axes=(0, 0, 0, 0))(
                keys,
                jnp.arange(n_results),
                bin_counts,
                bins,
            )

        return _random_neighbour

    return random_neighbour_decorator



"""
Graph/network utility functions
"""
from typing import Any, Tuple, Union

import chex
import jax
import jax.numpy as jnp


def sort_edges(x: chex.Array, *args) -> Union[chex.Array, Tuple[chex.Array, Any]]:
    """
    Sort graph edge indices and data for use in graph transformations

    Graph transformations expect graph edge indices to be sorted
    by the start index. Consequently, any data associated with
    edges should also be sorted in the same manner.

    Examples
    --------

    .. testsetup:: sort_edges

       import esquilax
       import jax

       k = jax.random.PRNGKey(101)
       n_nodes = 20
       n_edges = 10

    .. testcode:: sort_edges

       # Generate indices representing random graph edges
       edges = jax.random.choice(k, n_nodes, shape=(2, n_edges))
       sorted_edges = esquilax.utils.sort_edges(edges)

       # Include edge-weights
       edge_weights = jax.random.uniform(k, n_edges)
       sorted_edges, sorted_weights = esquilax.utils.sort_edges(
           edges, edge_weights
       )

    Parameters
    ----------
    x: jax.ndarray
        Array of edge indices in the shape :code:`[2, n_edges]`
    *args: Any
        Optional arguments representing edge data. These arguments
        can also be PyTrees which will then sort all members
        of the tree.

    Returns
    -------
    Sorted indices and sorted optional data
    """
    idxs = jnp.lexsort((x[1], x[0]))
    x = jnp.stack([x[0][idxs], x[1][idxs]])

    if args:
        args = jax.tree_util.tree_map(lambda a: a[idxs], args)
        return x, *args
    else:
        return x


def index_bins(indices: chex.Array, length: int) -> Tuple[chex.Array, chex.Array]:
    """
    Count indices frequencies and generate bins of contiguous indices

    This effectively generates a histogram of integer values, and
    corresponding start and end indices of the binned values if they
    are in sort order.

    Examples
    --------

    .. testsetup:: index_bins

       import esquilax
       import jax.numpy as jnp

       from esquilax.utils.graph import index_bins

    .. doctest:: index_bins

       >>> indices = jnp.array([0, 0, 1, 3, 3])
       >>> counts, bins = index_bins(indices, 5)
       >>> counts
       Array([2, 1, 0, 2, 0], dtype=int32)
       >>> bins
       Array([[0, 2],
              [2, 3],
              [3, 3],
              [3, 5],
              [5, 5]], dtype=int32)

    Parameters
    ----------
    indices: jax.ndarray
        1D array of integer indices
    length: int
        Number of indices to count, e.g. the
        number of bins to generate and count.

    Returns
    -------
    [jax.ndarray, jax.ndarray]
        Count/frequency of each index, and bins boundaries
    """
    bin_counts = jnp.bincount(indices, length=length)
    cum_counts = jnp.cumsum(bin_counts)

    return bin_counts, jnp.column_stack([cum_counts - bin_counts, cum_counts])
